
Meet3D::usage = "Meet3D[a, b] gives the intersection of a and b; a and b can be (in either order) a line and a plane, a line and a sphere, a pair of planes, a plane and a sphere, or a pair of spheres. 
Meet3D[a, b, c], where a, b and c are planes, gives their intersection."

Meet3D[Line @ {p_, q_}, w_Plane] := Module[
	{pp, qq, t},
	pp = Project[w][p];
	qq = Project[w][q];
	If[pp == qq, Return[pp]];
	t = Distance[p, pp] Tan[Angle[pp - p, q - p]];
	pp + t Unit[qq - pp] ] /; 0 != (p - q) . PlaneNormal @ w 

Meet3D[w_Plane, l_Line] := Meet3D[l, w]

Meet3D[v:Plane @ {a_, b_, c_}, w:Plane @ {p_, q_, r_}] := Module[
	{ptf, qtf, rtf, pqparallel, qrparallel, rpparallel},
	(* See which of {p, q, r} lie in v. *)
	ptf = CoplanarQ[{a, b, c, p}];
	qtf = CoplanarQ[{a, b, c, q}]; 
	rtf = CoplanarQ[{a, b, c, r}];

	(* Two of {p, q, r} lie in v. *)
 	If[ptf && qtf, Return[Line @ {p, q}]]; 
 	If[qtf && rtf, Return[Line @ {q, r}]]; 
 	If[rtf && ptf, Return[Line @ {r, p}]];

	(* See which of {p, q}, {q, r}, {r, p} are parallel to v. *)
	n = PlaneNormal @ w;
	pqparallel = 0 == (q - p) . n;
	qrparallel = 0 == (r - q) . n;
	rpparallel = 0 == (p - r) . n;

	(* One of {p, q, r} lies in v. 
	Find a line to intersect v that is not that one point. *)
	If[
		ptf, 
		If[qrparallel, 
			Return @ Line[{p, Meet3D[Line[{q, .5(r + p)}], v]}],
			Return @ Line[{p, Meet3D[Line[{q, r}], v]}]
		]
	];
	If[
		qtf, 
		If[rpparallel, 
			Return @ Line[{q, Meet3D[Line[{r, .5(p + q)}], v]}],
			Return @ Line[{p, Meet3D[Line[{r, p}], v]}]
		]
	];
	If[
		rtf, 
		If[pqparallel, 
			Return @ Line[{r, Meet3D[Line[{p, .5(q + r)}], v]}],
			Return @ Line[{r, Meet3D[Line[{p, q}], v]}]
		]
	];

	(* None of {p, q, r} lie on v. 
	Find two lines to intersect v. *)
	If[
		pqparallel, 
		Return @ 
		Line[{Meet3D[Line[{p, r}], v], Meet3D[Line[{q, r}], v]}]
	];
	If[
		qrparallel, 
		Return @ 
		Line[{Meet3D[Line[{q, p}], v], Meet3D[Line[{r, p}], v]}]
	];
	Line[{Meet3D[Line[{r, q}], v], Meet3D[Line[{p, q}], v]}]
] /; Cross[PlaneNormal @ v, PlaneNormal @ w] != $o 

Meet3D[u_Plane, v_Plane, w_Plane] := Meet3D[u, Meet3D[v, w]]

Meet3D[l:Line @ {p_, q_}, Sphere[e_, r_]] := Module[
	{a, b, c, d, t1, t2},
	a = (q - p) . (q - p);
	b = (q - p) . (e - p);
	c = (e - p) . (e - p) - r ^ 2;
	(* We solve the quadratic a x ^ 2 - 2 b x + c == 0. *)
	d = N @ Sqrt[b ^ 2 - a c];
	t1 = (b - d) / a;
	t2 = (b + d) / a;
	{p + (q - p) t1, p + (q - p) t2} ] /; Distance[e, l] <= r 

Meet3D[s_Sphere, l_Line] := Meet3D[l, s]

Meet3D[w:Plane @ {f_, g_, h_}, s:Sphere[c1_, r_]] := Module[{c, x, y, d, e},
	c = (c1 + Reflect[w] @ c1) / 2;
	{x, y} = Take[Complement[
		{f, g, h}, {c}, SameTest -> (Norm[#1-#2] < 10^-4 &)], 2];
	d = Last @ Meet3D[Line @ {c, x}, s]; 
	e = Last @ Meet3D[Line @ {c, y}, s];
	Circle3D[c, d, e] ] /; Distance[c1, w] < r

Meet3D[s_Sphere, w_Plane] := Meet3D[w, s]

Meet3D[Sphere[c1_, r1_], Sphere[c2_, r2_]] := Module[{C1, C2, x, c, d, e},
	C1 = Circle[$o, r1];
	C2 = Circle[Distance[c1, c2] $i, r2];
	x = First @ Meet[C1, C2];
	(* The line cd is perpendicular to the line 
		through the three centers. *)
	c = c1 + x[[1]] Unit[c2 - c1];
	d = c + x[[2]] Unit @ Cross[c1 - c, 
		If[Cross[c1 - c, $i - c] == $o, Cross[c1 - c, $j - c],
		Cross[c1 - c, $i - c]]];
	e = Rotate[N[Pi/2], Line@{c1, c2}][d];
	Circle3D[c, d, e] 
	] /; Abs[r1 - r2] <= Distance[c1, c2] <= r1 + r2 

